/**
 *      
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version. 
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * The Original Code is Copyright (C) 2006 by Bernd Breitenbach.
 * All rights reserved.
 *
 * The Original Code is: all of this file.
 *
 */

/**
 * \file mazegimp.c
 * \author Bernd Breitenbach
 * \remarks
 *      bb - Aug 27, 2006: Created.
 */

#include <string.h>
#include <stdlib.h>

#include <gtk/gtk.h>

#include <libgimp/gimp.h>
#include <libgimp/gimpui.h>

#include "mazeintern.h"

#define PLUG_IN_NAME "plug_in_maze5"


#ifdef ENABLE_NLS
#define GETTEXT_PACKAGE "gimp20-maze5"

#include <libintl.h>
#define _(String) gettext (String)
#ifdef gettext_noop
#    define N_(String) gettext_noop (String)
#else
#    define N_(String) (String)
#endif

/* #define bind_textdomain_codeset(Domain, Codeset) (Domain) */
#define INIT_I18N()     G_STMT_START{                           \
  bindtextdomain (GETTEXT_PACKAGE, gimp_locale_directory ());   \
  bind_textdomain_codeset (GETTEXT_PACKAGE, "UTF-8");           \
  textdomain (GETTEXT_PACKAGE);                                 \
}G_STMT_END

#else /* ! ENABLE_NLS */
#define _(String) (String)
#define N_(String) (String)
#endif


#include "mazelib.c"
#include "mazestruct.c"


enum
{
  SIZE_KEEP=0,
  SIZE_SCALE=1,
  SIZE_CROP=2,
};  


struct 
{
  int                   first_time;
  maze_parameter_t      para;
  int                   carver_idx;
  int                   struct_idx;
  GimpRGB               maze_color;
  GimpRGB               sol_color;
  GimpRGB               no_color;
  int                   image_size;
  int                   seed_init;
  int                   use_mask;
  render_parameter_t    render_para;
} lvals={
  .first_time     =TRUE,
  .para={
    .columns      =10,
    .rows         =10,
    .entrance     =NO_DIRECTION,
    .exit         =NO_DIRECTION,
    .goal         ={{0.,0.,0.}},
    .structure    =&ortho_struct,
    .seed         =-1,
    .windingness  =0.5,
    .construct    =NULL,
    .carver       =&newRandomBacktracker2D,
    .renderer     =NULL,
    .render_idx   =RENDER_NONE,
    .mask         =NULL,
  },
  .carver_idx=0,
  .struct_idx=0,
  .image_size=SIZE_KEEP,
  .seed_init=FALSE,
  .use_mask=FALSE,
  .render_para={
    .drawno        =FALSE,
    .draw_marks    =TRUE,
    .draw_solution =FALSE,
    .fuzzy         =FALSE,
    .amplitude     =0.05,
    .bordersize    =16.,
    .n_fuzzy       =8,
  }
};

typedef struct _combo_box_item
{
  gchar *name;
  union 
  {
    void  *ptr;
    int   no;
  };
} combo_box_item_t;

static combo_box_item_t maze_types[]={
  {ortho_struct.name,{.ptr=&ortho_struct}},
  {hexa_struct.name,{.ptr=&hexa_struct}},
  {tria_struct.name,{.ptr=&tria_struct}},
  {cairo_struct.name,{.ptr=&cairo_struct}},
  {hexa_square_struct.name,{.ptr=&hexa_square_struct}},
  {octa_square_struct.name,{.ptr=&octa_square_struct}},
  {rhombic_struct.name,{.ptr=&rhombic_struct}},
  {hexa_rhombic_struct.name,{.ptr=&hexa_rhombic_struct}},
  {triangle_square_struct.name,{.ptr=&triangle_square_struct}},
};

static combo_box_item_t algorithm_types[]={
  {"Recursive Backtracker",{.ptr=&newRandomBacktracker2D}},
  {"Prim's Algorithm",{.ptr=&newPrimCarver2D}},
};

static combo_box_item_t size_types[]={
  {"Keep",{.no=0}},
  {"Scale",{.no=1}},
  {"Crop",{.no=2}},
};

static combo_box_item_t direction_types[]={
  {_("Border"),{.no=-1}},
  {_("North"),{.no=0}},
  {_("West"),{.no=1}},
  {_("South"),{.no=2}},
  {_("East"),{.no=3}},
};


static void query (void);
static void run   (const gchar      *name,
                   gint              nparam,
                   const GimpParam  *param,
                   gint             *nretvals,
                   GimpParam       **retvals);

static void maze (GimpDrawable *drawable);

static gboolean maze_dialog (GimpDrawable *drawable);

#define WITHIN(a, b, c) ((((a) <= (b)) && ((b) <= (c))) ? TRUE : FALSE)

GimpPlugInInfo PLUG_IN_INFO =
{
  NULL,  /* init_proc  */
  NULL,  /* quit_proc  */
  query, /* query_proc */
  run,   /* run_proc   */
};

MAIN ()

static void
query (void)
{
  static GimpParamDef args[] =
  {
    { GIMP_PDB_INT32,    "run_mode",   "Interactive, non-interactive" },
    { GIMP_PDB_IMAGE,    "image",      "The Image"                    },
    { GIMP_PDB_DRAWABLE, "drawable",   "The Drawable"                 },
    { GIMP_PDB_INT32,    "columns",    "No of Columns"   },
    { GIMP_PDB_INT32,    "rows",       "No of Rows"       },
    { GIMP_PDB_INT32,    "type",       "Maze type ("
                                                   "0 = Orthognoal, "
                                                   "1 = Hexagonal, "
                                                   "2 = Triagonal, "
                                                   "3 = Cairo style, "
                                                   "4 = Hexagons-Squares, "
                                                   "5 = Octagons-Squares, "
                                                   "6 = Rhombic, "
                                                   "7 = Hexagonal-Rhombic, "
                                                   "8 = Triangles-Squares)"},
    { GIMP_PDB_INT32,    "algorithm",   "Algorithm ("
                                                   "0 = Recursive Backtracker, "
                                                   "1 = Prim's Algorithm)"},
    { GIMP_PDB_INT32,    "seed",       "Random Seed"},
    { GIMP_PDB_INT32,    "entrance",   "Side of entrance ("
                                                   "-1 = Border,"
                                                   " 0 = North,"
                                                   " 1 = West,"
                                                   " 2 = South,"
                                                   " 3 = East)"},
    { GIMP_PDB_INT32,    "exit",       "Side of exit ("
                                                   "-1 = Border,"
                                                   " 0 = North,"
                                                   " 1 = West,"
                                                   " 2 = South,"
                                                   " 3 = East)"},
    { GIMP_PDB_INT32,    "border",      "Border width in pixel (0 = No Border)"},
    { GIMP_PDB_DRAWABLE, "mask",        "The mask drawable (-1 for no mask)" },

    { GIMP_PDB_COLOR,    "maze_color"    , "Maze Color"},
    { GIMP_PDB_INT32,    "draw_marks"    , "Draw Entrance/Exit Marks (0 = No, 1=Yes)"},
    { GIMP_PDB_INT32,    "draw_solution" , "Draw Solution (0 = No, 1=Yes)"},
    { GIMP_PDB_COLOR,    "solution_color", "Solution Color"},
    { GIMP_PDB_INT32,    "draw_numbers"  , "Draw Cell Numbers (0 = No, 1=Yes)"},
    { GIMP_PDB_COLOR,    "number_color"  , "Number Color"},
  };

  gimp_install_procedure (PLUG_IN_NAME,
                          "Draw a maze",
                          "none yet",
                          "Bernd Breitenbach",
                          "Bernd Breitenbach",
                          "2006",
                          N_("_Maze5..."),
                          "RGB*, GRAY*",
                          GIMP_PLUGIN,
                          G_N_ELEMENTS (args), 0,
                          args, NULL);

  gimp_plugin_menu_register (PLUG_IN_NAME, "<Image>/Filters/Render/Pattern");
}

static void
run (const gchar      *name,
     gint              nparams,
     const GimpParam  *param,
     gint             *nreturn_vals,
     GimpParam       **return_vals)
{
  static GimpParam   values[1];
  GimpPDBStatusType  status = GIMP_PDB_SUCCESS;
  GimpDrawable      *drawable;
  int                imgID;
  *nreturn_vals = 1;
  *return_vals  = values;

  values[0].type          = GIMP_PDB_STATUS;
  values[0].data.d_status = status;

#ifdef ENABLE_NLS
  INIT_I18N ();
#endif

  imgID=param[1].data.d_int32;
  drawable = gimp_drawable_get (param[2].data.d_drawable);


  switch (param[0].data.d_int32)
    {

    case GIMP_RUN_INTERACTIVE:
      gimp_get_data (PLUG_IN_NAME, &lvals);
      lvals.para.mask=NULL;

      if (! maze_dialog(drawable))
        return;

    break;

    case GIMP_RUN_NONINTERACTIVE:
      if (nparams != 18)
        {
          status = GIMP_PDB_CALLING_ERROR;
        }
      else
        {
          lvals.para.columns    = param[3].data.d_int32;
          lvals.para.rows       = param[4].data.d_int32;
          lvals.struct_idx      = param[5].data.d_int32;
	  lvals.para.structure  = (cell_structure_t*)maze_types[lvals.struct_idx].ptr;
          lvals.carver_idx      = param[6].data.d_int32;
          lvals.para.carver     = algorithm_types[lvals.carver_idx].ptr;
          lvals.para.seed       = param[7].data.d_int32;
          lvals.para.entrance   = param[8].data.d_int32;
          lvals.para.exit       = param[9].data.d_int32;
	  lvals.render_para.bordersize= param[10].data.d_int32;
	  if (param[11].data.d_int32>=0) {
	    lvals.use_mask  =TRUE;
	    lvals.para.mask =(maze_2d_mask_t*)gimp_drawable_get(param[11].data.d_int32);
	  } else {
	    lvals.use_mask  =FALSE;
	  }
          lvals.maze_color               = param[12].data.d_color;
	  lvals.render_para.draw_marks   = param[13].data.d_int32;
	  lvals.render_para.draw_solution= param[14].data.d_int32;
          lvals.sol_color                = param[15].data.d_color;
	  lvals.render_para.drawno       = param[16].data.d_int32;
          lvals.no_color                 = param[17].data.d_color;
	}
      break;

    case GIMP_RUN_WITH_LAST_VALS:
      gimp_get_data (PLUG_IN_NAME, &lvals);
      lvals.para.mask=NULL;
      break;

    default:
      break;
    }

  if (status == GIMP_PDB_SUCCESS)
    {
      gimp_image_undo_group_start(param[1].data.d_image);
      maze (drawable);
      gimp_image_undo_group_end(param[1].data.d_image);

      gimp_set_data (PLUG_IN_NAME, &lvals, sizeof (lvals));
      values[0].data.d_status = status;
    }
}




static void toggle_update(GtkToggleButton *button,int *data)
{
  *data=gtk_toggle_button_get_active(button);
}

static void color_update(GimpColorButton *button,GimpRGB *color)
{
  gimp_color_button_get_color(button,color);
}


int maze2DMaskQuery(maze_2d_mask_t *mask,double x,double y)
{
  int retVal=TRUE;
  if (lvals.use_mask) {
    int bpp;
    guint8 *val;
    GimpDrawable *drawable=(GimpDrawable*)lvals.para.mask;
    bpp=drawable->bpp;
    if (x<0.) x=0.;
    else if (x>1) x=1.;
    if (y<0.) y=0.;
    else if (y>1) y=1.;
    val=gimp_drawable_get_pixel(drawable->drawable_id,
				(int)(drawable->width*x),(int)(drawable->height*y),&bpp);
    if (bpp>=3) {
      retVal=val[0]>0x80 && val[1]>0x80&& val[2]>0x80;
    } else {
      retVal=val[0]>0x80;
    }
  }
  return retVal;
}



static void paintStroke(GimpDrawable *drawable,stroke_t *stroke) 
{
  gdouble segment[STROKE_N_POINTS(stroke)*4];
  int i;
  int idx=0;
  for(i=0;i<STROKE_N_POINTS(stroke)-1;i++) {
    if (STROKE_TYPE(stroke,i+1)==STROKE_LINE) {
      segment[idx++]=STROKE_X(stroke,i);
      segment[idx++]=STROKE_Y(stroke,i);
      segment[idx++]=STROKE_X(stroke,i+1);
      segment[idx++]=STROKE_Y(stroke,i+1);
    } else {
      gimp_paintbrush_default(drawable->drawable_id,idx,segment);
      gimp_progress_update(i/(double)STROKE_N_POINTS(stroke));
      idx=0;
    }
  }
  if (idx>0) {
    gimp_paintbrush_default(drawable->drawable_id,idx,segment);
  }
}
 
static void drawNumbers(GimpDrawable *drawable,maze_2d_t *laby,render_parameter_t *para) 
{
  int i;
  double xsize= para->xcellsize;
  double ysize= para->ycellsize;
  cell_2d_t *cell;
  char buf[32];
  char *font=gimp_context_get_font();
  int imgID=gimp_drawable_get_image(drawable->drawable_id);
  gint width,height,ascent,descent;
  gdouble fsize=xsize/3.;
  int id;

  for(i=0;i<MAZE_N_CELLS(laby);i++) {
    cell=MAZE_CELL(laby,i);
    snprintf(buf,31,"%d",CELL_NO(cell));
    gimp_text_get_extents_fontname(buf,fsize,GIMP_PIXELS,
				   font,&width,&height,&ascent,&descent);
    id=gimp_text_fontname(imgID,drawable->drawable_id,
		       CELL_X(cell)*xsize+para->xoff-width/2.,
		       CELL_Y(cell)*ysize+para->yoff-height/2,
		       buf,0,TRUE,fsize,GIMP_PIXELS,font);
    gimp_floating_sel_anchor(id); 
  }
}

static GimpDrawable *newLayer(gchar *name,GimpDrawable *drawable)
{
  int imgID=gimp_drawable_get_image(drawable->drawable_id);
  
  int layerID=gimp_layer_new(imgID,name,drawable->width,drawable->height,
			 gimp_drawable_type_with_alpha(drawable->drawable_id),
			 100.,GIMP_NORMAL_MODE);
  gimp_drawable_fill(layerID,GIMP_TRANSPARENT_FILL);
  gimp_image_add_layer(imgID,layerID,-1);
  return gimp_drawable_get(layerID);
}



static void
maze (GimpDrawable *drawable)
{
  maze_2d_t *laby;
  carver_2d_t *carver;
  stroke_t *stroke;
  maze_2d_dim_t dim;
  int cborder;
  double border;

  gimp_context_push();

  drawable=newLayer("Maze5",drawable);

  lvals.render_para.width=drawable->width;
  lvals.render_para.height=drawable->height;

  laby=new2DMaze(&lvals.para);
  carver=lvals.para.carver(&lvals.para);
  carveMaze2D(carver,laby);

  dim=maze2DDimension(laby);

  border = lvals.para.tileable ? 0. : lvals.render_para.bordersize;
  lvals.render_para.xcellsize=(drawable->width-2*border)/(dim.xmax-dim.xmin);
  lvals.render_para.ycellsize=(drawable->height-2*border)/(dim.ymax-dim.ymin);
  if (lvals.image_size!=SIZE_SCALE) {
    if (lvals.render_para.xcellsize<lvals.render_para.ycellsize) {
      lvals.render_para.ycellsize=lvals.render_para.xcellsize;
    } else {
      lvals.render_para.xcellsize=lvals.render_para.ycellsize;
    }
  }
  lvals.render_para.xoff=border-(lvals.render_para.xcellsize*dim.xmin);
  lvals.render_para.yoff=border-(lvals.render_para.ycellsize*dim.ymin);

  gimp_context_set_foreground(&lvals.maze_color);
  stroke=renderMaze2D(laby,&lvals.render_para);
  gimp_progress_init(_("Drawing Maze..."));
  paintStroke(drawable,stroke);
  gimp_progress_update(100.);
  freeStroke2D(stroke);

  if (lvals.render_para.draw_marks) {
    stroke=renderEntryExit(laby,MAZE_ENTRANCE(laby),&lvals.render_para);
    paintStroke(drawable,stroke);
    freeStroke2D(stroke);

    stroke=renderEntryExit(laby,MAZE_EXIT(laby),&lvals.render_para);
    paintStroke(drawable,stroke);
    freeStroke2D(stroke);
  }

  if (lvals.render_para.draw_solution) {
    gimp_context_set_foreground(&lvals.sol_color);
    stroke=renderMaze2DSolution(laby,&lvals.render_para);
    gimp_progress_init(_("Drawing Solution..."));
    paintStroke(newLayer("Maze5 Solution",drawable),stroke);
    gimp_progress_update(100.);
    freeStroke2D(stroke);
  }

  if (lvals.render_para.drawno) {
    gimp_context_set_foreground(&lvals.no_color);
    drawNumbers(newLayer("Maze5 Numbers",drawable),laby,&lvals.render_para);
  }
  gimp_drawable_flush (drawable);
  gimp_drawable_update (drawable->drawable_id,0,0,drawable->width,drawable->height);

  if (lvals.image_size==SIZE_CROP) {
    if (lvals.para.tileable) {
      cborder=1;
    } else {
      cborder=2*lvals.render_para.bordersize;
    }
    gimp_image_crop(gimp_drawable_get_image(drawable->drawable_id),
		    MIN(lvals.render_para.xcellsize*(dim.xmax-dim.xmin)+cborder,drawable->width),
		    MIN(lvals.render_para.ycellsize*(dim.ymax-dim.ymin)+cborder,drawable->height),0,0);
  }

  freeMaze2D(laby);

  gimp_displays_flush ();
  gimp_context_pop();

}


/************************************************************************/

static void new_box_section(gchar *title,GtkBox *box,int rows,GtkWidget **table,GtkWidget **frame)
{

  *table = gtk_table_new (2, rows, FALSE);
  gtk_table_set_col_spacings (GTK_TABLE (*table), 12);
  gtk_table_set_row_spacings (GTK_TABLE (*table), 12);

  *frame=gimp_frame_new(title);
  gtk_container_add(GTK_CONTAINER(*frame),*table);
  gtk_widget_show(*table);

  gtk_box_pack_start(box,*frame,FALSE,FALSE,0);
  gtk_widget_show(*frame);
}

/************************************************************************/

#define C_BUTTON_SIZE (16)

static GtkWidget *new_check_button(gchar *title,int *val_ptr,GtkWidget *table,int col,int row)
{
  GtkWidget *widget=gtk_check_button_new();
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(widget),*val_ptr);
  g_signal_connect(widget, "toggled",G_CALLBACK (toggle_update),val_ptr);
  gimp_table_attach_aligned(GTK_TABLE(table),col,row,title,1.0,0.5,widget,1,TRUE);
  return widget;
}

static GtkWidget *new_color_button(gchar *title,GimpRGB *val_ptr,GtkWidget *table,int col,int row)
{
  GtkWidget *widget=gimp_color_button_new(title,2*C_BUTTON_SIZE,C_BUTTON_SIZE,
					  val_ptr,GIMP_COLOR_AREA_SMALL_CHECKS);
  g_signal_connect(widget, "color-changed",G_CALLBACK (color_update),val_ptr);
  gimp_table_attach_aligned(GTK_TABLE(table),col,row,title,1.0,0.5,widget,1,TRUE);
  return widget;
}


static GtkWidget *new_int_spin_button(gchar *title,int *val_ptr,int lower,int upper,
				      GtkWidget *table,int col,int row)
{
  GtkObject *adj;
  GtkWidget *widget=gimp_spin_button_new(&adj,*val_ptr,lower,upper,1,1,0,1,0);
  g_signal_connect (adj, "value_changed",G_CALLBACK(gimp_int_adjustment_update),val_ptr);
  gimp_table_attach_aligned(GTK_TABLE(table),col,row,title,1.0,0.5,widget,1,TRUE);
  return widget;
}

static GtkWidget *new_double_spin_button(gchar *title,double *val_ptr,
					 double lower,double upper,double incr,
					 GtkWidget *table,int col,int row)
{
  int digits=0;

  if (incr<1.) {
    digits=(int)(log(incr)/log(0.1)+1-1E-6);
  }
  GtkObject *adj;
  GtkWidget *widget=gimp_spin_button_new(&adj,*val_ptr,lower,upper,incr,1,0,1,digits);
  g_signal_connect (adj, "value_changed",G_CALLBACK(gimp_double_adjustment_update),val_ptr);
  gimp_table_attach_aligned(GTK_TABLE(table),col,row,title,1.0,0.5,widget,1,TRUE);
  return widget;
}

static GtkWidget *new_seed_button(gchar *title,int *seed_ptr,int *seed_init,
				  GtkWidget *table,int col,int row)
{
  GtkWidget *widget=gimp_random_seed_new((guint32*)seed_ptr,seed_init);
  gimp_table_attach_aligned(GTK_TABLE(table),col,row,title,1.0,0.5,widget,1,TRUE);
  return widget;
}





static void combo_box_ptr_update(GtkComboBox *box,int *idx_ptr)
{
  void **val_ptr=(void **)g_object_get_data(G_OBJECT(box),"maze5_val_ptr");
  combo_box_item_t *items=(combo_box_item_t *)g_object_get_data(G_OBJECT(box),"maze5_values");
  *idx_ptr=gtk_combo_box_get_active(box);
  *val_ptr=items[*idx_ptr].ptr;
}

static void combo_box_value_update(GtkComboBox *box,int *idx_ptr)
{
  combo_box_item_t *items=(combo_box_item_t *)g_object_get_data(G_OBJECT(box),"maze5_values");
  *idx_ptr=items[gtk_combo_box_get_active(box)].no;
}

#if 0
static void combo_box_int_update(GtkComboBox *box,int *idx_ptr)
{
  *idx_ptr=gtk_combo_box_get_active(box);
}
#endif

static void combo_box_drawable_update(GimpIntComboBox *box,GimpDrawable **drawable)
{
  gint32  drawable_id;

  gimp_int_combo_box_get_active(box, &drawable_id);
  *drawable=gimp_drawable_get(drawable_id);
}


static GtkWidget *new_combo_box_int(gchar *title,int n_items,combo_box_item_t *items,
				int *idx_ptr,GtkWidget *table,int col,int row)
{
  int i;
  GtkWidget *widget = gimp_int_combo_box_new(items[0].name,0,NULL);
  for (i=1;i<n_items;i++) {
    gimp_int_combo_box_append(GIMP_INT_COMBO_BOX(widget),
			      GIMP_INT_STORE_VALUE,items[i].no,
			      GIMP_INT_STORE_LABEL,items[i].name,
			      -1);
  }
  for(i=0;i<n_items;i++) {
    if (items[i].no==*idx_ptr) {
      gtk_combo_box_set_active(GTK_COMBO_BOX(widget),i);
      break;
    }
  }
  g_object_set_data(G_OBJECT(widget),"maze5_values",items);

  gimp_int_combo_box_connect(GIMP_INT_COMBO_BOX(widget),*idx_ptr,
			     G_CALLBACK(combo_box_value_update),idx_ptr);
  gimp_table_attach_aligned(GTK_TABLE(table),col,row,title,1.0,0.5,widget,1,TRUE);
  return widget;

}


static GtkWidget *new_combo_box_ptr(gchar *title,int n_items,combo_box_item_t *items,
				int *idx_ptr,void *val_ptr,
				  GtkWidget *table,int col,int row)
{
  int i;
  GtkWidget *widget = gimp_int_combo_box_new(items[0].name,0,NULL);
  for (i=1;i<n_items;i++) {
    gimp_int_combo_box_append(GIMP_INT_COMBO_BOX(widget),
			      GIMP_INT_STORE_VALUE,i,
			      GIMP_INT_STORE_LABEL,items[i].name,
			      -1);
  }
  g_object_set_data(G_OBJECT(widget),"maze5_values",items);
  g_object_set_data(G_OBJECT(widget),"maze5_val_ptr",val_ptr);

  gimp_int_combo_box_connect(GIMP_INT_COMBO_BOX(widget),*idx_ptr,
			     G_CALLBACK(combo_box_ptr_update),idx_ptr);
  gimp_table_attach_aligned(GTK_TABLE(table),col,row,title,1.0,0.5,widget,1,TRUE);
  return widget;
}


static gboolean drawable_constrain (gint32   image_id,
				    gint32   drawable_id,
				    gpointer data)
{
  return (gimp_drawable_is_rgb(drawable_id) || gimp_drawable_is_gray(drawable_id));
}


static GtkWidget *new_combo_box_drawable(gchar *title,GimpDrawable **drawable,
					 GtkWidget *table,int col,int row)
{
  int id=0;
  GtkWidget *widget=gimp_drawable_combo_box_new (drawable_constrain, NULL);

  if (*drawable!=NULL) {
    id=(*drawable)->drawable_id;
  }
  gimp_int_combo_box_connect (GIMP_INT_COMBO_BOX (widget),id,
                              G_CALLBACK(combo_box_drawable_update),
                              drawable);
  gimp_table_attach_aligned(GTK_TABLE(table),col,row,title,1.0,0.5,widget,1,TRUE);
  return widget;
}

/************************************************************************/

static void _sensitive_on_toggle_true_callback(GtkWidget *dependend,GtkToggleButton *button)
{
  gtk_widget_set_sensitive(dependend,gtk_toggle_button_get_active(button));
}


static void sensitive_on_toggle_true(GtkWidget *dependend,GtkWidget *widget)
{
  GtkToggleButton *button=GTK_TOGGLE_BUTTON(widget);
  _sensitive_on_toggle_true_callback(dependend,button);
  g_signal_connect_swapped(button, "toggled",
			   G_CALLBACK(_sensitive_on_toggle_true_callback),dependend);
}


static void _sensitive_on_toggle_false_callback(GtkWidget *dependend,GtkToggleButton *button)
{
  gtk_widget_set_sensitive(dependend,!gtk_toggle_button_get_active(button));
}


static void sensitive_on_toggle_false(GtkWidget *dependend,GtkWidget *widget)
{
  GtkToggleButton *button=GTK_TOGGLE_BUTTON(widget);
  _sensitive_on_toggle_false_callback(dependend,button);
  g_signal_connect_swapped(button, "toggled",
			   G_CALLBACK(_sensitive_on_toggle_false_callback),dependend);
}


/************************************************************************/



static gboolean
maze_dialog (GimpDrawable *drawable)
{
  GtkWidget *dialog;
  GtkWidget *main_vbox;
  GtkWidget *widget;
  GtkWidget *table;
  GtkWidget *frame;
  GtkWidget *solution;
  GtkWidget *numbers;
  GtkWidget *fuzzy;
  GtkWidget *tileable;
  GtkWidget *mask;

  gboolean   run;



  if (lvals.first_time) {
    gimp_context_get_foreground(&lvals.maze_color);
    gimp_context_get_foreground(&lvals.sol_color);
    gimp_context_get_foreground(&lvals.no_color);
    lvals.first_time=FALSE;
  }

  gimp_ui_init ("maze", TRUE);

  dialog = gimp_dialog_new (_("Maze"), "maze",
                            NULL, 0,
                            gimp_standard_help_func, "plug-in-maze5",
                            GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
                            GTK_STOCK_OK,     GTK_RESPONSE_OK,
                            NULL);

  main_vbox = gtk_vbox_new (FALSE, 12);
  gtk_container_set_border_width (GTK_CONTAINER (main_vbox), 12);
  gtk_container_add (GTK_CONTAINER (GTK_DIALOG (dialog)->vbox), main_vbox);
  gtk_widget_show (main_vbox);

  
  /** creation **/
  new_box_section("Maze Creation",GTK_BOX (main_vbox),7,&table,&frame);


  new_combo_box_ptr(_("Type"),sizeof(maze_types)/sizeof(combo_box_item_t),maze_types,
		    &lvals.struct_idx,&lvals.para.structure,table,0,0);

  new_combo_box_ptr(_("Algorithm"),sizeof(algorithm_types)/sizeof(combo_box_item_t),algorithm_types,
		    &lvals.carver_idx,&lvals.para.carver,table,1,0);
  
  new_int_spin_button(_("Columns"),&lvals.para.columns,2,1000,table,0,1);
  new_int_spin_button(_("Rows"),&lvals.para.rows,2,1000,table,1,1);

  new_seed_button(_("Seed"),&lvals.para.seed,&lvals.seed_init,table,0,3);

  tileable=new_check_button(_("Tileable"),&lvals.para.tileable,table,1,3);


  mask=new_check_button(_("Mask"),&lvals.use_mask,table,0,4);
  widget=new_combo_box_drawable(_("Map"),(GimpDrawable**)&lvals.para.mask,table,1,4);
  sensitive_on_toggle_true(widget,mask);

  new_combo_box_int(_("Entrance"),sizeof(direction_types)/sizeof(combo_box_item_t),
			   direction_types,&lvals.para.entrance,table,0,5);
  new_combo_box_int(_("Exit"),sizeof(direction_types)/sizeof(combo_box_item_t),
			   direction_types,&lvals.para.exit,table,1,5);


  /** rendering **/
  new_box_section("Drawing",GTK_BOX (main_vbox),6,&table,&frame);

  widget=new_double_spin_button(_("Border Width"),&lvals.render_para.bordersize,0.,1000.,1.,table,0,0);
  sensitive_on_toggle_false(widget,tileable);

  new_color_button(_("Maze Color"),&lvals.maze_color,table,1,0);

  solution=new_check_button(_("Draw Solution"),&lvals.render_para.draw_solution,table,0,1);
  widget=new_color_button(_("Solution Color"),&lvals.sol_color,table,1,1);
  sensitive_on_toggle_true(widget,solution);

  numbers=new_check_button(_("Cell Numbers"),&lvals.render_para.drawno,table,0,2);
  widget=new_color_button(_("Number Color"),&lvals.no_color,table,1,2);
  sensitive_on_toggle_true(widget,numbers);

  new_check_button(_("Draw Entrance/Exit"),&lvals.render_para.draw_marks,table,0,3);

  new_combo_box_int(_("Size"),sizeof(size_types)/sizeof(combo_box_item_t),size_types,
		    &lvals.image_size,table,0,4);
  
  fuzzy=new_check_button(_("Fuzzy Lines"),&lvals.render_para.fuzzy,table,0,5);
  widget=new_int_spin_button(_("No of Samples"),&lvals.render_para.n_fuzzy,2,50,table,0,6);
  sensitive_on_toggle_true(widget,fuzzy);
  widget=new_double_spin_button(_("Amplitude"),&lvals.render_para.amplitude,0.,2.,0.01,table,1,6);
  sensitive_on_toggle_true(widget,fuzzy);
                    
  gtk_widget_show(table);
  gtk_widget_show(dialog);

  run = (gimp_dialog_run (GIMP_DIALOG (dialog)) == GTK_RESPONSE_OK);

  gtk_widget_destroy (dialog);

  return run;
}




