/**
 *      
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version. 
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * The Original Code is Copyright (C) 2006 by Bernd Breitenbach.
 * All rights reserved.
 *
 * The Original Code is: all of this file.
 *
 */


/**
 * \file mazelib.h
 * \author Bernd Breitenbach
 * \remarks
 *      bb - Aug 13, 2006: Created.
 */


#ifndef _MAZELIB_H
#define _MAZELIB_H


#ifndef FALSE
#define FALSE (0)
#endif

#ifndef TRUE
#define TRUE (!FALSE)
#endif

typedef double coord_t;

typedef enum _direction
{
  GOAL =-2,
  NO_DIRECTION=-1,
  NORTH=0,
  WEST =1,
  SOUTH=2,
  EAST =3
} direction_t;


typedef enum _cell_state
{
  CELL_UNUSED   =0,
  CELL_USED     =1,
  CELL_OUT      =0xff
} cell_state_t;

typedef enum _cell_type
{
  TYPE_A=0,
  TYPE_B=1,
  TYPE_C=2,
  TYPE_D=3,
  TYPE_E=4,
  TYPE_F=5,
  TYPE_G=6,
  TYPE_H=7,
  TYPE_I=8,
  TYPE_J=9,
  TYPE_VALID=9000,
  TYPE_OFF = 9999,
  TYPE_NONE=99999,
} cell_type_t;


typedef enum _visit_type
{
  NOT_VISITED=0,
  VISITED    =1,
  FRONTIER   =2
} visit_t; 


typedef enum _wall_type
{
  WALL_OPEN          =0,      /* no wall */
  WALL_CLOSED        =1,      /* completely closed */
} wall_type_t;


typedef enum _laby_type
{
  MAZE_POLY_2D       =0,
  MAZE_THETA_2D      =1,
  MAZE_FREE          =255,
} laby_type_t;

typedef enum _carver_type
{
  CARVER_BACKTRACKER =0,
  CARVER_PRIM        =1,
} carver_type_t;


typedef enum _render_format
{
  RENDER_NONE =-1,
  RENDER_ASCII=0,
  RENDER_PDF  =1,
  RENDER_PNG  =2,
  RENDER_SVG  =3,
} render_format_t;

/**********************************************************************/

struct _render_parameter;
struct _maze_parameter;
struct _cell_structure;
struct _cell_2d;
struct _border;

/**********************************************************************/

typedef union _user_data
{
  void            *ptr_data;
  long            long_data;
} user_data_u;


typedef struct _cell_2d_conn
{
  wall_type_t     wall;
  struct _cell_2d *neighbour;
  int             neighbour_idx;
  user_data_u     user;
} cell_2d_conn_t;


typedef union _coords
{
  struct
  {
    coord_t x,y,z;
  } cartesic;
  struct
  {
    coord_t r,a,z;
  } polar;
} coords_u;


typedef struct _cell_2d
{
  cell_state_t    state;
  cell_type_t     type;
  int             no;
  coords_u        coords;
  int             n_neighbours;
  cell_2d_conn_t  *neighbours;
  struct _border  *border;
  user_data_u     user;
} cell_2d_t;


typedef int (*cell_predicate_f)(cell_2d_t *);

/**********************************************************************/

typedef enum _stroke_type
{
  STROKE_MOVE=0,
  STROKE_LINE=1,
} stroke_type_t;

typedef struct _single_stroke{
  stroke_type_t type;
  coords_u      point;
  cell_2d_t    *cell;
} single_stroke_t;

typedef struct _stroke
{
  int             n_points;
  int             max_points;
  single_stroke_t points[];
} stroke_t;

/**********************************************************************/


typedef struct _maze_2d
{
  laby_type_t type;
  struct _cell_structure *structure;  
  int        columns;
  int        rows;
  int        tileable;
  int        n_cells;
  cell_2d_t *cells;
  cell_2d_t *entrance;
  cell_2d_t *exit;
} maze_2d_t;


/**********************************************************************/

typedef struct _border
{
  coords_u point;
  int      neighbour_x;
  int      neighbour_y;
} border_t;


typedef struct _cell_structure
{
  char         name[32];
  int          rows;
  int          columns;
  coord_t      height;
  coord_t      width;
  coord_t     xminadd;
  coord_t     yminadd;
  coord_t     xmaxadd;
  coord_t     ymaxadd;
  coords_u    *center_points;
  cell_type_t *cell_types;
  int         *n_neighbours;
  border_t    *cell_coords[];
} cell_structure_t;


/**********************************************************************/

typedef struct _maze_2d_dim
{
  double xmin;
  double ymin;
  double xmax;
  double ymax;
} maze_2d_dim_t;


/**********************************************************************/

typedef struct _maze_2d_mask
{
  void *data;
} maze_2d_mask_t;

/**********************************************************************/

typedef struct _point_2d
{
  int       idx;
  cell_2d_t *cell;
} point_2d_t;


typedef struct _path_2d
{
  int n_steps;
  point_2d_t *steps;
} path_2d_t;

/**********************************************************************/

typedef struct _carver_2d
{
  carver_type_t type;
  maze_2d_t *laby;
  void       (*entrance)(maze_2d_t *,struct _carver_2d *);
  void       (*exit)(maze_2d_t *,struct _carver_2d *);
  void       (*carve)(struct _carver_2d *);
  cell_2d_t *current;
  struct _maze_parameter *para;
} carver_2d_t;

/**********************************************************************/

typedef void (*laby_render_f)(maze_2d_t *,struct _render_parameter *);

typedef struct _maze_parameter
{
  int              columns;
  int              rows;
  direction_t      entrance;
  direction_t      exit;
  coords_u         goal;
  cell_structure_t *structure;
  int              seed;
  double           windingness;
  int              tileable;
  maze_2d_t        *(*construct)(struct _maze_parameter *);
  carver_2d_t      *(*carver)(struct _maze_parameter *);
  laby_render_f    *renderer;
  int render_idx;
  maze_2d_mask_t   *mask;
} maze_parameter_t;

/**********************************************************************/

typedef struct _color
{
  double red;
  double green;
  double blue;
  double alpha;
} color_t;


typedef struct _render_parameter
{
  laby_render_f render;
  char   *filename;
  double width;
  double height;
  double xcellsize;
  double ycellsize;
  double bordersize;
  double xoff;
  double yoff;
  double linewidth;
  int    drawno;
  color_t foreground;
  color_t background;
  int    draw_marks;
  int    draw_solution;
  int    fuzzy;
  double amplitude;
  int    n_fuzzy;
  char   *solution_text;
  char   *copyright_text;
}render_parameter_t;


/**********************************************************************/

extern maze_2d_t *new2DMaze(maze_parameter_t *parameter);

extern void freeMaze2D(maze_2d_t *laby);

/**********************************************************************/

extern cell_2d_t *cell2D(maze_2d_t *laby,int col, int row);

/**********************************************************************/

extern maze_2d_t *new2DThetaLabyrinth(maze_parameter_t *parameter);

extern cell_2d_t *cell2DTheta(maze_2d_t *laby,int ring, int pos);

/**********************************************************************/

extern void renderPDF2D(maze_2d_t *laby,render_parameter_t *para);

extern void renderSVG2D(maze_2d_t *laby,render_parameter_t *para);

extern void renderPNG2D(maze_2d_t *laby,render_parameter_t *para);

/**********************************************************************/

extern void resetMaze2DUserData(maze_2d_t *laby);

extern int visitWall(cell_2d_t *cell,int idx);

extern int visitNeighbour(cell_2d_t *cell,int idx);

extern int visitCell(cell_2d_t *cell);


extern void setWall2D(cell_2d_t *cell,int idx,wall_type_t wall);

extern void tagWall2D(cell_2d_t *cell,int idx);


extern void makeOpening2D(cell_2d_t *cell,int idx);

extern wall_type_t wall2D(cell_2d_t *cell,int idx);

extern int isOpen2D(cell_2d_t *cell,int idx);

extern int isClosed2D(cell_2d_t *cell,int idx);

extern int isTagged2D(cell_2d_t *cell,int idx);

/**********************************************************************/

extern carver_2d_t *newRandomBacktracker2D(maze_parameter_t *parameter);

extern carver_2d_t *newPrimCarver2D(maze_parameter_t *parameter);

extern void freeCarver2D(carver_2d_t *carver);

/**********************************************************************/

extern void carveMaze2D(carver_2d_t *carver,maze_2d_t *laby);

/**********************************************************************/

extern maze_2d_mask_t *newMaze2DMask(char *filename);

extern int maze2DMaskQuery(maze_2d_mask_t *mask,double x,double y);

extern maze_2d_dim_t maze2DDimension(maze_2d_t *laby);

/**********************************************************************/

extern stroke_t *newStroke2D(int initsize);

extern stroke_t *resizeStroke2D(stroke_t *stroke,int newsize);

extern void freeStroke2D(stroke_t *stroke);

extern  stroke_t *appendStroke2DPoint(stroke_t *stroke,stroke_type_t type,double x,double y);

extern stroke_t *appendStroke2DCell(stroke_t *stroke,stroke_type_t type,double x,double y,cell_2d_t *cell);

extern stroke_t *renderMaze2D(maze_2d_t *laby,render_parameter_t *para);

extern stroke_t *renderMaze2DSolution(maze_2d_t *laby,render_parameter_t *para);

extern stroke_t *renderEntryExit(maze_2d_t *maze,cell_2d_t *cell,render_parameter_t *para);

extern void renderTextAlongStroke(maze_2d_t *maze,stroke_t *stroke,char *text);

/**********************************************************************/

#endif /* _MAZELIB_H */
