/*
 * @(#) $Id: waves-of-chaos.c 69 2005-12-28 23:18:39Z yeti $
 * This is a plugin for the GIMP.
 *
 * Copyright (C) 2004 Yeti (David Necas) <yeti@physics.muni.cz>.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 */

/*
 * This plug-in generates wavy tileable textures by solving a pair of
 * coupled differential equations.
 *
 * Please send patches, suggestions, and other invectives to me:
 * <yeti@physics.muni.cz>.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#else
/* To include libgimp/stdplugins-intl.h */
#  define GETTEXT_PACKAGE "gimp20"
#endif /* HAVE_CONFIG_H */

#include <stdlib.h>
#include <string.h>

#include <gtk/gtk.h>

#include <libgimp/gimp.h>
#include <libgimp/gimpui.h>
#include "libgimp/stdplugins-intl.h"

#define PROCEDURE_NAME   "waves_of_chaos"
#define DATA_KEY_VALS    "plug_in_waves_of_chaos"

enum {
    BITS_PER_SAMPLE   = 8,
    CHANNEL_MAX_VALUE = (1U << BITS_PER_SAMPLE) - 1,

    ENTRY_WIDTH =          75,
    PREVIEW_SIZE =         128,
    SCALE_WIDTH  =         201,

    /* How often progressbar is updated */
    PROGRESS_STEP = 20,
    PROGRESS_PREVIEW_STEP = 100
};

/* Empiric factor translating integration parameters to feature size */
#define FEATURE_SIZE_TO_H     0.576

enum {
    RESPONSE_REVERT = 1,
    RESPONSE_RESET
};

enum {
    COLUMN_QUANTITY,
    COLUMN_NAME
};

typedef enum {
    COLORING_METHOD_GRAY = 0,
    COLORING_METHOD_RGB  = 1
} ColoringMethod;

typedef enum {
    QUANTITY_VALUE_1           = 0,
    QUANTITY_VALUE_2           = 1,
    QUANTITY_TIME_DERIVATION_1 = 2,
    QUANTITY_TIME_DERIVATION_2 = 3,
    QUANTITY_DIFFUSION_SPEED_1 = 4,
    QUANTITY_DIFFUSION_SPEED_2 = 5,
    QUANTITY_DIFFUSION_X_1     = 6,
    QUANTITY_DIFFUSION_X_2     = 7,
    QUANTITY_DIFFUSION_Y_1     = 8,
    QUANTITY_DIFFUSION_Y_2     = 9,
    QUANTITY_CONST_0           = 10,
    QUANTITY_CONST_1_2         = 11,
    QUANTITY_CONST_1           = 12
} ChannelQuantity;

typedef struct {
    gfloat min;
    gfloat max;
} DataRange;

typedef struct {
    gdouble         feature_size;
    ColoringMethod  mode;
    ChannelQuantity gray;
    ChannelQuantity red;
    ChannelQuantity green;
    ChannelQuantity blue;
} PluginValues;

static void query                (void);
static void run                  (const gchar *name,
                                  gint nparams,
                                  const GimpParam *param,
                                  gint *nreturn_vals,
                                  GimpParam **return_vals);
static gint plugin_dialog        (void);
static void channel_value_changed(GtkWidget *widget,
                                  ChannelQuantity *valp);
static void type_changed         (GtkWidget *button,
                                  gpointer data);
static void feature_size_changed (gboolean *valid);
static void refresh_controls     (const PluginValues *new_pvals);
static void commit               (GimpDrawable *drawable,
                                  const gint channel);
static void commit_one_tile      (GimpPixelRgn *rgn,
                                  const gint channel);
static void execute              (GimpDrawable *drawable);
static void iterate              (gint width,
                                  gint height,
                                  gint iterations,
                                  gfloat h,
                                  gboolean initialize);
static void do_iter              (gint xres,
                                  gint yres,
                                  gfloat *data,
                                  gfloat h);

typedef struct {
    GtkObject *feature_size;
    GtkWidget *grayscale;
    GtkWidget *rgb;
    GtkWidget *update;
    GtkWidget *gray;
    GtkWidget *gray_label;
    GtkWidget *red;
    GtkWidget *red_label;
    GtkWidget *green;
    GtkWidget *green_label;
    GtkWidget *blue;
    GtkWidget *blue_label;
} PluginControls;

static GimpPlugInInfo PLUG_IN_INFO = {
    NULL,  /* init_proc  */
    NULL,  /* quit_proc  */
    query, /* query_proc */
    run,   /* run_proc   */
};

static const PluginValues pvals_defaults =
{
    10.0,                 /* feature_size */
    COLORING_METHOD_GRAY, /* mode         */
    0,                    /* gray         */
    0,                    /* red          */
    1,                    /* green        */
    2,                    /* blue         */
};

static PluginControls pctrl;
static PluginValues pvals; /* = pvals_defaults; */
static PluginValues pvals_old; /* for revert */

static GRand *rng = NULL;
static GtkWidget *preview_image = NULL;
static GdkPixbuf *preview = NULL;
static GtkWidget *preview_progressbar = NULL;
static gboolean preview_mode;
static gboolean is_rgb;
static GimpDrawable *drawable;
static gboolean data_valid = FALSE;

static gint ix1, iy1, ix2, iy2;     /* Selected image size. */
static gint bpp, has_alpha, alpha;
static glong progress;
static gdouble max_progress;
static ChannelQuantity channel_map[3];
static DataRange data_range[5];
static gfloat *pdata;


MAIN()

static void
query(void)
{
    static GimpParamDef args[] = {
        { GIMP_PDB_INT32,    "run_mode", "Interactive, non-interactive" },
        { GIMP_PDB_IMAGE,    "image", "Input image (unused)" },
        { GIMP_PDB_DRAWABLE, "drawable", "Input drawable" },
        { GIMP_PDB_FLOAT,    "feature_size", "Characteristic feature size" },
        { GIMP_PDB_INT32,    "mode", "Coloring mode: { GRAYSCALE (0), RGB (1) }" },
        { GIMP_PDB_INT32,    "gray", "Quantity to use for grayscale (0..12)" },
        { GIMP_PDB_INT32,    "red", "Quantity to use for red (0..12)" },
        { GIMP_PDB_INT32,    "green", "Quantity to use for green (0..12)" },
        { GIMP_PDB_INT32,    "blue", "Quantity to use for blue (0..12)" },
    };

    gimp_install_procedure
        (DATA_KEY_VALS,
         "Renderers a wavy chaotic pattern",
         "This plug-in renderes a snapshot of evolution of chaotic "
         "two-dimensional partial differential equation.",
         "Yeti",
         "Yeti",
         "Nov 2004",
         N_("<Image>/Filters/Render/_Waves of Chaos..."),
         "RGB*, GRAY*",
         GIMP_PLUGIN,
         G_N_ELEMENTS(args), 0, args, NULL);
}

static void
run(const gchar *name,
    gint nparams,
    const GimpParam  *param,
    gint *nreturn_vals,
    GimpParam  **return_vals)
{
    static GimpParam values[1];
    GimpRunMode run_mode;
    GimpPDBStatusType status = GIMP_PDB_SUCCESS;

    run_mode = param[0].data.d_int32;
	gchar *LOCALEDIR = _glib_get_locale_dir (); //samj
    bindtextdomain(GETTEXT_PACKAGE, LOCALEDIR);
#ifdef HAVE_BIND_TEXTDOMAIN_CODESET
    bind_textdomain_codeset(GETTEXT_PACKAGE, "UTF-8");
#endif
    textdomain(GETTEXT_PACKAGE);

    *nreturn_vals = 1;
    *return_vals = values;

    values[0].type = GIMP_PDB_STATUS;
    values[0].data.d_status = status;

    drawable = gimp_drawable_get(param[2].data.d_drawable);
    is_rgb = gimp_drawable_is_rgb(drawable->drawable_id);
    has_alpha = gimp_drawable_has_alpha(drawable->drawable_id);

    pvals = pvals_defaults;
    rng = g_rand_new();
    switch (run_mode) {
        case GIMP_RUN_INTERACTIVE:
        gimp_get_data(DATA_KEY_VALS, &pvals);

        pvals_old = pvals;
        if (!plugin_dialog()) {
            gimp_drawable_detach(drawable);
            return;
        }
        break;

        case GIMP_RUN_NONINTERACTIVE:
        if (nparams != 9)
            status = GIMP_PDB_CALLING_ERROR;
        else {
            pvals.feature_size = (gint)param[3].data.d_float;
            pvals.mode = (gint)param[4].data.d_int32;
            pvals.gray = (gint)param[5].data.d_int32;
            pvals.red = (gint)param[6].data.d_int32;
            pvals.green = (gint)param[7].data.d_int32;
            pvals.blue = (gint)param[8].data.d_int32;
        }
        break;

        case GIMP_RUN_WITH_LAST_VALS:
        gimp_get_data(DATA_KEY_VALS, &pvals);
        break;

        default:
        break;
    }

    if (status == GIMP_PDB_SUCCESS) {
        if (is_rgb || gimp_drawable_is_gray(drawable->drawable_id)) {
            gimp_progress_init(_("Waves of Chaos..."));

            preview_mode = FALSE;
            execute(drawable);

            if (run_mode != GIMP_RUN_NONINTERACTIVE)
                gimp_displays_flush();

            if (run_mode == GIMP_RUN_INTERACTIVE)
                gimp_set_data(DATA_KEY_VALS, &pvals, sizeof(PluginValues));
        }
        else
            status = GIMP_PDB_EXECUTION_ERROR;
    }

    values[0].data.d_status = status;
    gimp_drawable_detach(drawable);
}

/* remove some source code redundancies {{{ */
static inline GtkWidget*
yeti_preview_new_with_progress(GtkWidget *hbox,
                               GimpDrawable *drawable,
                               GCallback callback,
                               GtkWidget **progressbar,
                               GtkWidget **button)
{
    GtkWidget *frame, *vbox, *image;
    GdkPixbuf *pixbuf;

    vbox = gtk_vbox_new(FALSE, 0);
    gtk_container_set_border_width(GTK_CONTAINER(vbox), 4);
    gtk_box_pack_start(GTK_BOX(hbox), vbox, FALSE, FALSE, 0);

    frame = gtk_frame_new(NULL);
    gtk_frame_set_shadow_type(GTK_FRAME(frame), GTK_SHADOW_IN);
    gtk_box_pack_start(GTK_BOX(vbox), frame, FALSE, FALSE, 0);

    pixbuf = gdk_pixbuf_new(GDK_COLORSPACE_RGB,
                            FALSE,
                            BITS_PER_SAMPLE,
                            PREVIEW_SIZE,
                            PREVIEW_SIZE);
    gdk_pixbuf_fill(pixbuf, 0);

    image = gtk_image_new_from_pixbuf(pixbuf);
    g_object_unref(pixbuf);
    gtk_container_add(GTK_CONTAINER(frame), image);

    *progressbar = gtk_progress_bar_new();
    gtk_widget_set_size_request(*progressbar, PREVIEW_SIZE, 8);
    gtk_box_pack_start(GTK_BOX(vbox), *progressbar, FALSE, FALSE, 0);

    *button = gtk_button_new_with_mnemonic(_("_Update Preview"));
    gtk_box_pack_start(GTK_BOX(vbox), *button, FALSE, FALSE, 0);
    g_signal_connect(G_OBJECT(*button), "clicked",
                     G_CALLBACK(callback), drawable);

    return image;
}

static inline GtkObject*
yeti_scale_entry_new_double(const gchar *name,
                            GtkWidget *table,
                            GtkSizeGroup *sgroup,
                            const gint row,
                            double *value,
                            const double min,
                            const double max,
                            const double step,
                            const double page,
                            const gint digits)
{
    GtkObject *adj;

    *value = CLAMP(*value, min, max);
    adj = gimp_scale_entry_new(GTK_TABLE(table), 0, row, name, SCALE_WIDTH, 0,
                               *value, min, max, step, page, digits,
                               TRUE, 0, 0, NULL, NULL);
    g_signal_connect(adj, "value-changed",
                     G_CALLBACK(gimp_double_adjustment_update), value);
    if (sgroup)
        gtk_size_group_add_widget(sgroup, GIMP_SCALE_ENTRY_SPINBUTTON(adj));

    return adj;
}
/* }}} */

static void
channel_quantity_combo_set_active(GtkWidget *combo,
                                  ChannelQuantity value)
{
    ChannelQuantity v;
    GtkTreeModel *model;
    GtkTreeIter iter;

    model = gtk_combo_box_get_model(GTK_COMBO_BOX(combo));
    gtk_tree_model_get_iter_first(model, &iter);
    do {
        gtk_tree_model_get(model, &iter, COLUMN_QUANTITY, &v, -1);
        if (v == value) {
            gtk_combo_box_set_active_iter(GTK_COMBO_BOX(combo), &iter);
            return;
        }
    } while (gtk_tree_model_iter_next(model, &iter));
}

static GtkWidget*
channel_quantity_combo_new(GtkWidget *table,
                           const gchar *name,
                           gint *row,
                           ChannelQuantity *valp,
                           GtkWidget **label)
{
    static const struct {
        const gchar *name;
        ChannelQuantity q;
    }
    values[] = {
        { N_("Value 1"),           QUANTITY_VALUE_1, },
        { N_("Value 2"),           QUANTITY_VALUE_2, },
        { N_("Time derivation 1"), QUANTITY_TIME_DERIVATION_1, },
        { N_("Time derivation 2"), QUANTITY_TIME_DERIVATION_2, },
        { N_("Diffusion speed 1"), QUANTITY_DIFFUSION_SPEED_1, },
        { N_("Diffusion speed 2"), QUANTITY_DIFFUSION_SPEED_2, },
        { N_("Diffusion-X 1"),     QUANTITY_DIFFUSION_X_1, },
        { N_("Diffusion-X 2"),     QUANTITY_DIFFUSION_X_2, },
        { N_("Diffusion-Y 1"),     QUANTITY_DIFFUSION_Y_1, },
        { N_("Diffusion-Y 2"),     QUANTITY_DIFFUSION_Y_2, },
        { N_("Constant 0"),        QUANTITY_CONST_0, },
        { N_("Constant 1/2"),      QUANTITY_CONST_1_2, },
        { N_("Constant 1"),        QUANTITY_CONST_1, },
    };
    GtkListStore *store;
    GtkTreeIter iter;
    GtkCellRenderer *renderer;
    GtkWidget *combo;
    guint i;

    store = gtk_list_store_new(2, G_TYPE_UINT, G_TYPE_STRING);
    for (i = 0; i < G_N_ELEMENTS(values); i++) {
        gtk_list_store_append(store, &iter);
        gtk_list_store_set(store, &iter,
                           COLUMN_QUANTITY, values[i].q,
                           COLUMN_NAME, _(values[i].name),
                           -1);
    }

    combo = gtk_combo_box_new_with_model(GTK_TREE_MODEL(store));
    gtk_combo_box_set_wrap_width(GTK_COMBO_BOX(combo), 1);
    renderer = gtk_cell_renderer_text_new();
    gtk_cell_layout_pack_start(GTK_CELL_LAYOUT(combo), renderer, TRUE);
    gtk_cell_layout_add_attribute(GTK_CELL_LAYOUT(combo), renderer,
                                  "text", COLUMN_NAME);
    channel_quantity_combo_set_active(combo, *valp);
    g_signal_connect(combo, "changed",
                     G_CALLBACK(channel_value_changed), valp);

    gtk_table_attach(GTK_TABLE(table), combo, 2, 3, *row, *row + 1,
                     GTK_EXPAND | GTK_FILL, 0, 2, 2);
    *label = gtk_label_new_with_mnemonic(name);
    gtk_misc_set_alignment(GTK_MISC(*label), 1.0, 0.5);
    gtk_table_attach(GTK_TABLE(table), *label, 1, 2, *row, *row + 1,
                     GTK_EXPAND | GTK_FILL, 0, 2, 2);
    gtk_label_set_mnemonic_widget(GTK_LABEL(*label), combo);

    (*row)++;

    return combo;
}

static gint
plugin_dialog(void)
{
    GtkWidget *dlg, *hbox, *vbox, *table;
    gint run, row;
    gboolean ok;

    preview_mode = TRUE;
    gimp_ui_init(PROCEDURE_NAME, TRUE);

    dlg = gimp_dialog_new(_("Waves of Chaos"), PROCEDURE_NAME,
                          NULL, 0,
                          gimp_standard_help_func,
                          "filters/waves_of_chaos.html",

                          _("Re_vert"), RESPONSE_REVERT,
                          GIMP_STOCK_RESET, RESPONSE_RESET,
                          GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
                          GTK_STOCK_OK, GTK_RESPONSE_OK,
                          NULL);

    hbox = gtk_hbox_new(FALSE, 8);
    gtk_container_set_border_width(GTK_CONTAINER(hbox), 6);
    gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dlg)->vbox), hbox, TRUE, TRUE, 0);

    preview_image = yeti_preview_new_with_progress(hbox, drawable,
                                                   G_CALLBACK(execute),
                                                   &preview_progressbar,
                                                   &pctrl.update);
    preview = gtk_image_get_pixbuf(GTK_IMAGE(preview_image));

    vbox = gtk_vbox_new(FALSE, 0);
    gtk_container_set_border_width(GTK_CONTAINER(vbox), 4);
    gtk_box_pack_start(GTK_BOX(hbox), vbox, TRUE, TRUE, 0);

    table = gtk_table_new(3, 1, FALSE);
    gtk_box_pack_start(GTK_BOX(vbox), table, FALSE, FALSE, 0);
    gtk_table_set_col_spacings(GTK_TABLE(table), 4);
    gtk_table_set_row_spacings(GTK_TABLE(table), 2);
    row = 0;

    pctrl.feature_size = yeti_scale_entry_new_double(_("_Feature size:"),
                                                     table, NULL, row,
                                                     &pvals.feature_size,
                                                     4.0, 40.0, 1.0, 4.0, 1);
    g_signal_connect_swapped(pctrl.feature_size, "value-changed",
                             G_CALLBACK(feature_size_changed), &data_valid);
    row++;

    table = gtk_table_new(4, 3, FALSE);
    gtk_box_pack_start(GTK_BOX(vbox), table, FALSE, FALSE, 0);
    gtk_table_set_col_spacings(GTK_TABLE(table), 4);
    gtk_table_set_row_spacings(GTK_TABLE(table), 2);
    row = 0;

    if (!is_rgb)
        pvals.mode = COLORING_METHOD_GRAY;

    pctrl.grayscale = gtk_radio_button_new_with_mnemonic(NULL,
                                                         _("Gr_ayscale"));
    gtk_table_attach(GTK_TABLE(table), pctrl.grayscale, 0, 1, row, row+1,
                     GTK_EXPAND | GTK_FILL, 0, 2, 2);
    g_signal_connect(pctrl.grayscale, "toggled",
                     G_CALLBACK(type_changed),
                     GUINT_TO_POINTER(COLORING_METHOD_GRAY));

    pctrl.gray = channel_quantity_combo_new(table, _("Value:"), &row,
                                            &pvals.gray, &pctrl.gray_label);

    pctrl.rgb = gtk_radio_button_new_with_mnemonic_from_widget
                                (GTK_RADIO_BUTTON(pctrl.grayscale), _("R_GB"));
    gtk_table_attach(GTK_TABLE(table), pctrl.rgb, 0, 1, row, row+1,
                     GTK_EXPAND | GTK_FILL, 0, 2, 2);
    if (is_rgb)
        g_signal_connect(pctrl.rgb, "toggled",
                         G_CALLBACK(type_changed),
                         GUINT_TO_POINTER(COLORING_METHOD_RGB));
    else
        gtk_widget_set_sensitive(pctrl.rgb, FALSE);

    pctrl.red = channel_quantity_combo_new(table, _("R_ed:"), &row,
                                            &pvals.red, &pctrl.red_label);
    pctrl.green = channel_quantity_combo_new(table, _("Gree_n:"), &row,
                                            &pvals.green, &pctrl.green_label);
    pctrl.blue = channel_quantity_combo_new(table, _("_Blue:"), &row,
                                            &pvals.blue, &pctrl.blue_label);

    refresh_controls(&pvals);
    gtk_widget_show_all(dlg);

    ok = FALSE;
    do {
        run = gimp_dialog_run(GIMP_DIALOG(dlg));
        switch (run) {
            case RESPONSE_REVERT:
            refresh_controls(&pvals_old);
            break;

            case RESPONSE_RESET:
            refresh_controls(&pvals_defaults);
            break;

            default:
            ok = TRUE;
            break;
        }
    } while (!ok);

    gtk_widget_destroy(dlg);

    return run == GTK_RESPONSE_OK;
}

static void
channel_value_changed(GtkWidget *widget,
                      ChannelQuantity *valp)
{
    GtkTreeIter active;

    gtk_combo_box_get_active_iter(GTK_COMBO_BOX(widget), &active);
    gtk_tree_model_get(gtk_combo_box_get_model(GTK_COMBO_BOX(widget)), &active,
                       COLUMN_QUANTITY, valp, -1);
}

static void
feature_size_changed(gboolean *valid)
{
    *valid = FALSE;
}

static void
type_changed(GtkWidget *button,
             gpointer data)
{
    gboolean sens;

    if (!gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(button)))
        return;

    pvals.mode = GPOINTER_TO_UINT(data);
    sens = (pvals.mode == COLORING_METHOD_RGB);
    gtk_widget_set_sensitive(pctrl.gray, !sens);
    gtk_widget_set_sensitive(pctrl.gray_label, !sens);
    gtk_widget_set_sensitive(pctrl.red, sens);
    gtk_widget_set_sensitive(pctrl.red_label, sens);
    gtk_widget_set_sensitive(pctrl.green, sens);
    gtk_widget_set_sensitive(pctrl.green_label, sens);
    gtk_widget_set_sensitive(pctrl.blue, sens);
    gtk_widget_set_sensitive(pctrl.blue_label, sens);
}

static void
refresh_controls(const PluginValues *new_pvals)
{
    /* we cannot reset some field and don't want to reset some others, so copy
       them one by one  */
    pvals.feature_size = new_pvals->feature_size;
    gtk_adjustment_set_value(GIMP_SCALE_ENTRY_SCALE_ADJ(pctrl.feature_size),
                             pvals.feature_size);

    pvals.gray = new_pvals->gray;
    channel_quantity_combo_set_active(pctrl.gray, pvals.gray);

    pvals.red = new_pvals->red;
    channel_quantity_combo_set_active(pctrl.red, pvals.red);

    pvals.green = new_pvals->green;
    channel_quantity_combo_set_active(pctrl.green, pvals.green);

    pvals.blue = new_pvals->blue;
    channel_quantity_combo_set_active(pctrl.blue, pvals.blue);

    /* XXX: this is silly, but I'm lazy to implement proper initial dialog
     * setup */
    pvals.mode = new_pvals->mode;
    if (pvals.mode == COLORING_METHOD_GRAY) {
        gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(pctrl.rgb), TRUE);
        gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(pctrl.grayscale), TRUE);
    }
    else {
        gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(pctrl.grayscale), TRUE);
        gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(pctrl.rgb), TRUE);
    }
}

/****************************************************************************/

static void
execute(GimpDrawable *drawable)
{
    gint n, i, chmax, iterations;

    if (preview_mode)
        gtk_widget_set_sensitive(pctrl.update, FALSE);

    /* compute sizes and other stuff of this kind */
    alpha = -1000; /* to cause a segfault ;-> */
    if (preview_mode) {
        ix1 = iy1 = 0;
        ix2 = gdk_pixbuf_get_width(preview);
        iy2 = gdk_pixbuf_get_height(preview);
        bpp = gdk_pixbuf_get_n_channels(preview);
        n = (ix2 - ix1)*(iy2 - iy1);
        if (!pdata)
            pdata = g_new(gfloat, 5*n);
    }
    else {
        gimp_drawable_mask_bounds(drawable->drawable_id,
                                  &ix1, &iy1, &ix2, &iy2);
        bpp = gimp_drawable_bpp(drawable->drawable_id);
        if (pdata)
            g_free(pdata);
        n = (ix2 - ix1)*(iy2 - iy1);
        pdata = g_new(gfloat, 5*n);
        if (has_alpha)
            alpha = bpp-1;
    }
    chmax = bpp;

    switch (pvals.mode) {
        case COLORING_METHOD_GRAY:
        channel_map[0] = pvals.gray;
        channel_map[1] = pvals.gray;
        channel_map[2] = pvals.gray;
        break;

        case COLORING_METHOD_RGB:
        channel_map[0] = pvals.red;
        channel_map[1] = pvals.green;
        channel_map[2] = pvals.blue;
        break;
    }

    if (preview_mode && data_valid)
        iterations = 200*(1 + pvals.feature_size/40);
    else
        iterations = 1000*(1 + pvals.feature_size/40);
    max_progress = iterations;

    progress = 0;
    iterate(ix2 - ix1, iy2 - iy1, iterations,
            FEATURE_SIZE_TO_H/pvals.feature_size,
            !(preview_mode && data_valid));

    for (i = 0; i < chmax; i++)
        commit(drawable, i);

    if (preview_mode) {
        gtk_widget_queue_draw(preview_image);
        gtk_widget_set_sensitive(pctrl.update, TRUE);
        /* http://bugzilla.gnome.org/show_bug.cgi?id=56070 workaround */
        gtk_widget_hide(pctrl.update);
        gtk_widget_show(pctrl.update);
    }
    else {
        gimp_drawable_flush(drawable);
        gimp_drawable_merge_shadow(drawable->drawable_id, TRUE);
        gimp_drawable_update(drawable->drawable_id,
                             ix1, iy1, (ix2 - ix1), (iy2 - iy1));
    }
}

/* move one-channel data into preview or drawable */
static void
commit(GimpDrawable *drawable, const gint channel)
{
    gpointer pr;
    GimpPixelRgn dest_rgn;

    if (preview_mode) {
        dest_rgn.x = dest_rgn.y = 0;
        dest_rgn.w = dest_rgn.h = PREVIEW_SIZE;
        dest_rgn.bpp = gdk_pixbuf_get_n_channels(preview);
        dest_rgn.rowstride = gdk_pixbuf_get_rowstride(preview);
        dest_rgn.data = gdk_pixbuf_get_pixels(preview);
        commit_one_tile(&dest_rgn, channel);
    }
    else {
        gimp_pixel_rgn_init(&dest_rgn, drawable,
                            ix1, iy1, (ix2 - ix1), (iy2 - iy1),
                            TRUE, TRUE);
        /* iterate by tiles */
        for (pr = gimp_pixel_rgns_register(1, &dest_rgn);
             pr != NULL;
             pr = gimp_pixel_rgns_process(pr)) {
            commit_one_tile(&dest_rgn, channel);
        }
    }
}

/* move one-channel data into one tile of drawable
   when channel == alpha, just fill it with CHANNEL_MAX_VALUE */
static void
commit_one_tile(GimpPixelRgn *rgn, const gint channel)
{
    guint x, y;
    guchar *p;

    if (channel == alpha) {
        for (y = rgn->y; y < rgn->y + rgn->h; y++) {
            for (x = rgn->x; x < rgn->x + rgn->w; x++) {
                p = rgn->data + (y - rgn->y)*rgn->rowstride
                    + (x - rgn->x)*rgn->bpp;
                p[channel] = CHANNEL_MAX_VALUE;
            }
        }
    }
    else {
        gfloat *data = pdata + channel_map[channel]*(iy2 - iy1)*(ix2 - ix1);
        gfloat min = data_range[channel_map[channel]].min;
        gfloat max = data_range[channel_map[channel]].max;
        gfloat r;

        for (y = rgn->y; y < rgn->y + rgn->h; y++) {
            for (x = rgn->x; x < rgn->x + rgn->w; x++) {
                p = rgn->data + (y - rgn->y)*rgn->rowstride
                    + (x - rgn->x)*rgn->bpp;
                r = data[(y - iy1)*(ix2 - ix1) + x - ix1];
                r = (r - min)/(max - min);
                r = CLAMP(r, 0.0, 1.0);
                p[channel] = CHANNEL_MAX_VALUE*r;
            }
        }
    }
}

static void
compute_extra_quantity(gint width,
                       gint height,
                       gfloat *data,
                       ChannelQuantity quantity)
{
    gfloat *rdata, x, y;
    gint i, j;

    switch (quantity) {
        case QUANTITY_DIFFUSION_SPEED_1:
        case QUANTITY_DIFFUSION_SPEED_2:
        if (quantity == QUANTITY_DIFFUSION_SPEED_1)
            rdata = pdata;
        else
            rdata = pdata + width*height;

        for (i = 0; i < height; i++) {
            gint ip, im;

            ip = ((i + 1) % height) * width;
            im = ((i + height - 1) % height) * width;
            for (j = 0; j < width; j++) {
                gint jp, jm;

                jp = (j + 1) % width;
                jm = (j + width - 1) % width;
                x = rdata[i*width + jp] - rdata[i*width + jm];
                y = rdata[ip + j] - rdata[im + j];
                data[i*width + j] = sqrt(x*x + y*y);
            }
        }
        break;

        case QUANTITY_DIFFUSION_X_1:
        case QUANTITY_DIFFUSION_X_2:
        if (quantity == QUANTITY_DIFFUSION_X_1)
            rdata = pdata;
        else
            rdata = pdata + width*height;

        for (i = 0; i < height; i++) {
            for (j = 0; j < width; j++) {
                gint jp, jm;

                jp = (j + 1) % width;
                jm = (j + width - 1) % width;
                data[i*width + j] = rdata[i*width + jp] - rdata[i*width + jm];
            }
        }
        break;

        case QUANTITY_DIFFUSION_Y_1:
        case QUANTITY_DIFFUSION_Y_2:
        if (quantity == QUANTITY_DIFFUSION_Y_1)
            rdata = pdata;
        else
            rdata = pdata + width*height;

        for (i = 0; i < height; i++) {
            gint ip, im;

            ip = ((i + 1) % height) * width;
            im = ((i + height - 1) % height) * width;
            for (j = 0; j < width; j++)
                data[i*width + j] = rdata[j + ip] - rdata[j + im];
        }
        break;

        case QUANTITY_CONST_0:
        for (i = 0; i < width*height; i++)
            data[i] = 0.0;
        break;

        case QUANTITY_CONST_1_2:
        for (i = 0; i < width*height; i++)
            data[i] = 0.5;
        break;

        case QUANTITY_CONST_1:
        for (i = 0; i < width*height; i++)
            data[i] = 1.0;
        break;

        default:
        g_assert_not_reached();
        break;
    }
}

static DataRange
normalize(gint n, gfloat *data)
{
    DataRange dr;
    gfloat avg, rms, max, min;
    gint i;

    min = G_MAXFLOAT;
    max = -G_MAXFLOAT;
    for (i = 0; i < n; i++) {
        if (data[i] < min)
            min = data[i];
        if (data[i] > max)
            max = data[i];
    }

    if (min == max) {
        dr.min = 0.0;
        dr.max = 1.0;
        return dr;
    }

    avg = rms = 0.0;
    for (i = 0; i < n; i++)
        avg += data[i];
    avg /= n;
    for (i = 0; i < n; i++)
        rms += (data[i] - avg)*(data[i] - avg);
    rms = sqrt(rms/n);

    dr.min = MAX(avg - 3.0*rms, min);
    dr.max = MIN(avg + 3.0*rms, max);
    return dr;
}

/* remix channel map to use existing data, but make space for new data, and
 * keep first two intact */
static void
remap_channels(gint width, gint height)
{
    gint over4;
    gboolean need1dt, need2dt;
    ChannelQuantity i, j, k;

    need1dt = need2dt = FALSE;
    over4 = 0;
    if (channel_map[0] >= 4)
        over4++;
    if (channel_map[1] >= 4
        && channel_map[1] != channel_map[0])
        over4++;
    if (channel_map[2] >= 4
        && channel_map[2] != channel_map[0]
        && channel_map[2] != channel_map[1])
        over4++;

    need1dt = channel_map[0] == 2 || channel_map[1] == 2 || channel_map[3] == 2;
    need2dt = channel_map[0] == 3 || channel_map[1] == 3 || channel_map[3] == 3;

    switch (over4) {
        case 0:
        break;

        case 1:
        i = MAX(channel_map[0], channel_map[1]);
        i = MAX(i, channel_map[2]);
        compute_extra_quantity(width, height, pdata + 4*width*height, i);
        for (j = 0; j < 3; j++) {
            if (channel_map[j] == i)
                channel_map[j] = 4;
        }
        break;

        case 2:
        i = MAX(channel_map[0], channel_map[1]);
        i = MAX(i, channel_map[2]);
        if (channel_map[0] >= 4 && channel_map[0] != i)
            j = channel_map[0];
        else if (channel_map[1] >= 4 && channel_map[1] != i)
            j = channel_map[1];
        else
            j = channel_map[2];
        compute_extra_quantity(width, height, pdata + 4*width*height, i);
        k = need1dt ? 3 : 2;
        compute_extra_quantity(width, height, pdata + k*width*height, j);
        if (channel_map[0] == j)
            channel_map[0] = k;
        if (channel_map[1] == j)
            channel_map[1] = k;
        if (channel_map[2] == j)
            channel_map[2] = k;
        for (j = 0; j < 3; j++) {
            if (channel_map[j] == i)
                channel_map[j] = 4;
        }
        break;

        case 3:
        for (i = 0; i < 3; i++) {
            compute_extra_quantity(width, height, pdata + (i+2)*width*height,
                                   channel_map[i]);
            channel_map[i] = i+2;
        }
        break;

        default:
        g_assert_not_reached();
        break;
    }

    for (i = 0; i < 4 + (over4 != 0); i++)
        data_range[i] = normalize(width*height, pdata + i*width*height);
}

static void
iterate(gint width,
        gint height,
        gint iterations,
        gfloat h,
        gboolean initialize)
{
    gint i, n;

    n = width*height;
    if (initialize) {
        for (i = 0; i < 2*n; i++)
            pdata[i] = g_rand_double_range(rng, -1.0, 1.0);
    }

    for (i = 0; i < iterations; i++) {
        do_iter(width, height, pdata, h);
        progress++;
        if (preview_mode) {
            if (!(i % PROGRESS_PREVIEW_STEP)) {
                gtk_progress_bar_set_fraction
                                       (GTK_PROGRESS_BAR(preview_progressbar),
                                        progress/max_progress);
                while (gtk_events_pending())
                    gtk_main_iteration_do(FALSE);
            }
        }
        else {
            if (!(i % PROGRESS_STEP))
                gimp_progress_update(progress/max_progress);
        }
    }

    remap_channels(width, height);

    if (preview_mode) {
        gtk_progress_bar_set_fraction(GTK_PROGRESS_BAR(preview_progressbar),
                                      0.0);
        while (gtk_events_pending())
            gtk_main_iteration_do(FALSE);
    }

    data_valid = TRUE;
}

static void
do_iter(gint xres, gint yres, gfloat *data, gfloat h)
{
    gfloat *c0, *c1, *r0, *r1;
    gfloat mu0h, mu1h;
    gfloat p = 1.12;
    gfloat q = -1.4;
    gfloat q0 = 0.3;
    gfloat dt = 1.0/40;
    gint i, j, ix, ixp, ixm, jp, jm;

    c0 = data;
    c1 = data + xres*yres;
    r0 = data + 2*xres*yres;
    r1 = data + 3*xres*yres;
    mu0h = 0.001/h/h;
    mu1h = 0.0001/h/h;

    for (i = 0; i < yres; i++) {
        ix = i*xres;
        if (G_UNLIKELY(i == yres - 1))
            ixp = 0;
        else
            ixp = ix + xres;
        if (G_UNLIKELY(i == 0))
            ixm = (yres - 1)*xres;
        else
            ixm = ix - xres;
        for (j = 0; j < xres; j++) {
            gfloat c0lap, c1lap, cx0, cx1;

            if (G_UNLIKELY(j == xres - 1))
                jp = 0;
            else
                jp = j + 1;
            if (G_UNLIKELY(j == 0))
                jm = xres - 1;
            else
                jm = j - 1;
            cx0 = c0[ix + j];
            cx1 = c1[ix + j];
            /* mixed rectangular-diagonal laplaceans */
            c0lap = c0[ix + jp] + c0[ix + jm] + c0[ixp + j] + c0[ixm + j]
                    + 0.25*(c0[ixp + jp] + c0[ixp + jm]
                            + c0[ixm + jp] + c0[ixm + jm])
                    - 5.0*cx0;
            c1lap = c1[ix + jp] + c1[ix + jm] + c1[ixp + j] + c1[ixm + j]
                    + 0.25*(c1[ixp + jp] + c1[ixp + jm]
                            + c1[ixm + jp] + c1[ixm + jm])
                    - 5.0*cx1;
            r0[ix + j] = q0*cx0/(1 + cx0*cx0) + q*cx1 + mu0h*c0lap;
            r1[ix + j] = p*cx0*(1 + cx1*cx1) + mu1h*c1lap;
        }
    }
    for (i = 0; i < yres; i++) {
        for (j = 0; j < xres; j++) {
            c0[i*xres + j] += dt*r0[i*xres + j];
            c1[i*xres + j] += dt*r1[i*xres + j];
        }
    }
}

